let currentInputs={};
let handlers=[];
let config;
let hud;

const toAngleMagnitude=(x, y) => {
	return {
		angle: ((Math.atan2(x, y)+2*Math.PI)%(2*Math.PI))/Math.PI,
		magnitude: Math.hypot(x, y)
	};
};

const handleAngleMagnitude=(x, y, threshold=0, fn=null, clearBuffer=false) => {
	const {angle, magnitude}=toAngleMagnitude(x, y);

	if(magnitude>threshold) {
		let inputs=currentInputs;
		if(angle>.25 && angle <.75) inputs.right=true;
		else if(angle>.75 && angle<1.25) inputs.up=true;
		else if(angle>1.25 && angle<1.75) inputs.left=true;
		else inputs.down=true;

		if(clearBuffer) inputs.clearBuffer=true;
		if(fn) fn(angle, magnitude);
	}
}

const handleCrosspad=(() => {
	const ns='http://www.w3.org/2000/svg';
	const cross=document.createElementNS(ns, 'svg');
	cross.classList.add('crosspadOverlay');
	cross.setAttribute('width', 1000);
	cross.setAttribute('height', 1000);
	let dr=document.createElementNS(ns, 'line');
	dr.setAttribute('x1', 0);
	dr.setAttribute('y1', 0);
	dr.setAttribute('x2', 1000);
	dr.setAttribute('y2', 1000);
	dr.setAttribute('stroke', 'black');
	cross.appendChild(dr);
	let dl=document.createElementNS(ns, 'line');
	dl.setAttribute('x1', 1000);
	dl.setAttribute('y1', 0);
	dl.setAttribute('x2', 0);
	dl.setAttribute('y2', 1000);
	dl.setAttribute('stroke', 'black');
	cross.appendChild(dl);

	const fn=e =>
		handleAngleMagnitude(
			e.touches[0].clientX-window.innerWidth/2,
			e.touches[0].clientY-window.innerHeight/2,
			0,
			null,
			!config.touchscreen.buffer
		);
	return {
		touchstart: fn,
		touchmove: fn,
		init: () => hud.appendChild(cross),
		fini: () => hud.removeChild(cross)
	};
})();

const handleKeyboard={
	keydown: e => {
		let inputs=currentInputs;
		if(e.key=='ArrowUp') inputs.up=true;
		else if(e.key=='ArrowDown') inputs.down=true;
		else if(e.key=='ArrowLeft') inputs.left=true;
		else if(e.key=='ArrowRight') inputs.right=true;

		if(!config.keyboard.buffer) inputs.clearBuffer=true;
	}
};

const handleJoystick=(() => {
	let center={
		x: 0,
		y: 0
	};
	return {
		touchstart: e => {
			center.x=e.touches[0].clientX;
			center.y=e.touches[0].clientY;
		},
		touchmove: e =>
			handleAngleMagnitude(
				e.touches[0].clientX-center.x,
				e.touches[0].clientY-center.y,
				config.touchscreen.deadzone,
				null,
				!config.touchscreen.buffer
			)
	}
})();

const handleSwipe=(() => {
	let center={
		x: 0,
		y: 0
	};
	let resetCenter=e => {
		center.x=e.touches[0].clientX;
		center.y=e.touches[0].clientY;
	};
	return {
		touchstart: resetCenter,
		touchmove: e =>
			handleAngleMagnitude(
				e.touches[0].clientX-center.x,
				e.touches[0].clientY-center.y,
				config.touchscreen.deadzone,
				() => resetCenter(e),
				!config.touchscreen.buffer
			)
	}
})();

const handleGamepads={
	frame: () => {
		const gp=navigator.getGamepads()[0];
		let inputs=currentInputs;
		if(!gp || !gp.axes) return;

		handleAngleMagnitude(
			gp.axes[0],
			gp.axes[1],
			config.gamepad.deadzone,
			null,
			!config.gamepad.buffer
		);
	}
};

const handleEvent=(type, evt) => {
	for(let handler of handlers) {
		let fn=handler[type];
		if(fn) fn(evt);
	}
};

const enableHandler=handler => {
	if(!handlers.includes(handler)) {
		handlers.push(handler);
		if(handler.init) handler.init();
	}
};
const disableHandler=handler => {
	let idx=handlers.indexOf(handler);
	if(idx!=-1) {
		handlers.splice(idx, 1);
		if(handler.fini) handler.fini();
	}
};

const updateConfig=cfg =>
	config=cfg;
const setHud=elem =>
	hud=elem;

const clear=() =>
	Object
		.keys(currentInputs)
		.forEach(key => delete currentInputs[key]);

for(let type of ['keydown', 'touchstart', 'touchmove']) {
	window.addEventListener(type, handleEvent.bind(null, type));
}

return module.exports={
	inputs: currentInputs,
	clear,
	enableHandler, disableHandler,
	framefn: handleEvent.bind(null, 'frame'),
	availableHandlers: {
		keyboard: handleKeyboard,
		gamepad: handleGamepads,
		touchscreenCrosspad: handleCrosspad,
		touchscreenJoystick: handleJoystick,
		touchscreenSwipe: handleSwipe
	},
	updateConfig, setHud
};
